import logging

__all__ = []

# ---------------------------------------------------------

__all__.append("configure_logging")
def configure_logging(logger_name, level):
    """
    Configure logging format and level.
    """
    streamHandler = logging.StreamHandler()
    formatter = logging.Formatter(
        "[%(name)s] %(levelname)s: %(message)s"
    )
    streamHandler.setFormatter(formatter)
    logger = logging.getLogger(logger_name)
    logger.setLevel(level)
    logger.addHandler(streamHandler)
    logger.debug("Logging level set to DEBUG")

    return logger

# ---------------------------------------------------------

__all__.append("add_general_parameters")
def add_general_parameters(parser):
    """
    Add General Options used by all ssm-* tools.
    """
    group_general = parser.add_argument_group('General Options')
    group_general.add_argument('--profile', '-p', dest='profile', type=str, help='Configuration profile from ~/.aws/{credentials,config}')
    group_general.add_argument('--region', '-g', dest='region', type=str, help='Set / override AWS region.')
    group_general.add_argument('--verbose', '-v', action='store_const', dest='log_level', const=logging.INFO, default=logging.WARN, help='Increase log_level level')
    group_general.add_argument('--debug', '-d', action='store_const', dest='log_level', const=logging.DEBUG, help='Increase log_level level')
    group_general.add_argument('--help', '-h', action="help", help='Print this help and exit')

    return group_general

# ---------------------------------------------------------

__all__.append("bytes_to_human")
def bytes_to_human(size):
    """
    Convert Bytes to more readable units
    """
    units = [ 'B', 'kB', 'MB', 'GB', 'TB' ]
    unit_idx = 0    # Start with Bytes
    while unit_idx < len(units)-1:
        if size < 2048:
            break
        size /= 1024.0
        unit_idx += 1
    return size, units[unit_idx]

# ---------------------------------------------------------

__all__.append("seconds_to_human")
def seconds_to_human(seconds, decimal=3):
    """
    Convert seconds to HH:MM:SS[.SSS]

    If decimal==0 only full seconds are used.
    """
    secs = int(seconds)
    fraction = seconds - secs

    mins = int(secs / 60)
    secs = secs % 60

    hours = int(mins / 60)
    mins = mins % 60

    ret = f"{hours:02d}:{mins:02d}:{secs:02d}"
    if decimal:
        fraction = int(fraction * (10**decimal))
        ret += f".{fraction:0{decimal}d}"

    return ret



#!/usr/bin/env python3

import os
import re
import logging
import botocore.credentials
import botocore.session
import boto3

logger = logging.getLogger("ssm-session")

class InstanceResolver():
    def __init__(self, args):
        # aws-cli compatible MFA cache
        cli_cache = os.path.join(os.path.expanduser('~'),'.aws/cli/cache')

        # Construct boto3 session with MFA cache
        session = boto3.session.Session(profile_name=args.profile, region_name=args.region)
        session._session.get_component('credential_provider').get_provider('assume-role').cache = botocore.credentials.JSONFileCache(cli_cache)

        # Create boto3 clients from session
        self.ssm_client = session.client('ssm')
        self.ec2_client = session.client('ec2')

    def get_list(self):
        def _try_append(_list, _dict, _key):
            if _key in _dict:
                _list.append(_dict[_key])

        items = {}

        # List instances from SSM
        logger.debug("Fetching SSM inventory")
        inventory = self.ssm_client.get_inventory()
        for entity in inventory["Entities"]:
            try:
                content = entity['Data']['AWS:InstanceInformation']["Content"][0]

                # At the moment we only support EC2 Instances
                assert content["ResourceType"] == "EC2Instance"

                # Ignore Terminated instances
                if content.get("InstanceStatus") == "Terminated":
                    logger.debug("Ignoring terminated instance: %s", entity)
                    continue

                # Add to the list
                instance_id = content['InstanceId']
                items[instance_id] = {
                    "InstanceId": instance_id,
                    "HostName": content.get("ComputerName"),
                }
                logger.debug("Added instance: %s: %r", instance_id, items[instance_id])
            except (KeyError, ValueError):
                logger.debug("SSM inventory entity not recognised: %s", entity)
                continue

        # Add attributes from EC2
        reservations = self.ec2_client.describe_instances(InstanceIds=list(items.keys()))
        for reservation in reservations['Reservations']:
            for instance in reservation['Instances']:
                instance_id = instance['InstanceId']
                if not instance_id in items:
                    continue

                # Find instance IPs
                items[instance_id]['Addresses'] = []
                _try_append(items[instance_id]['Addresses'], instance, 'PrivateIpAddress')
                _try_append(items[instance_id]['Addresses'], instance, 'PublicIpAddress')

                # Find instance name from tag Name
                items[instance_id]['InstanceName'] = ""
                for tag in instance['Tags']:
                    if tag['Key'] == 'Name':
                        items[instance_id]['InstanceName'] = tag['Value']

                logger.debug("Updated instance: %s: %r", instance_id, items[instance_id])
        return items

    def print_list(self):
        hostname_len = 0
        instname_len = 0

        items = self.get_list().values()

        if not items:
            logger.warning("No instances registered in SSM!")
            return

        items = list(items)
        items.sort(key=lambda x: x.get('InstanceName') or x.get('HostName'))

        for item in items:
            hostname_len = max(hostname_len, len(item['HostName']))
            instname_len = max(instname_len, len(item['InstanceName']))

        for item in items:
            print(f"{item['InstanceId']}   {item['HostName']:{hostname_len}}   {item['InstanceName']:{instname_len}}   {' '.join(item['Addresses'])}")

    def resolve_instance(self, instance):
        # Is it a valid Instance ID?
        if re.match('^i-[a-f0-9]+$', instance):
            return instance

        # It is not - find it in the list
        instances = []

        items = self.get_list()
        for instance_id in items:
            item = items[instance_id]
            if instance.lower() in [item['HostName'].lower(), item['InstanceName'].lower()] + item['Addresses']:
                instances.append(instance_id)

        if not instances:
            return None

        if len(instances) > 1:
            logger.warning("Found %d instances for '%s': %s", len(instances), instance, " ".join(instances))
            logger.warning("Use INSTANCE_ID to connect to a specific one")
            quit(1)

        # Found only one instance - return it
        return instances[0]




import time
import logging
import pexpect

class SsmTalker():
    def __init__(self, instance_id, profile, region, logger_name):
        self._instance_id = instance_id
        self._logger = logging.getLogger(logger_name)
        self.connect(instance_id, profile, region)

    def connect(self, instance_id, profile, region):
        extra_args = ""
        if profile:
            extra_args += f"--profile {profile} "
        if region:
            extra_args += f"--region {region} "
        command = f'aws {extra_args} ssm start-session --target {instance_id}'
        self._logger.debug(f"Spawning: {command}")
        self._child = pexpect.spawn(command, echo=False, encoding='utf-8', timeout=10)
        #self._child.logfile_read = sys.stderr
        self._logger.debug(f"PID: {self._child.pid}")

        self.wait_for_prompt()
        self._logger.debug(f"{self._child.before.strip()}")
        self.shell_prompt = self._child.after

        # Turn off input echo
        self._child.sendline('stty -echo')
        self.wait_for_prompt()

        # Change to home directory (SSM session starts in '/')
        self._child.sendline('cd')
        self.wait_for_prompt()

    def exit(self):
        self._logger.debug("Closing session")
        self._child.sendcontrol('c')
        time.sleep(0.5)
        self._child.sendline('exit')
        try:
            self._child.expect(['Exiting session', pexpect.EOF])
        except (OSError, pexpect.exceptions.EOF):
            pass

    def wait_for_prompt(self):
        """
        As of now a typical SSM prompt is 'sh-4.2$ '
        """
        self._child.expect('sh.*\$ $')





import pathlib

from setuptools import setup, find_packages

import ssm_tools

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

SCRIPTS=[
    'ssm-session',
    'ssm-copy',
    'ssm-tunnel',
]

setup(
    name="aws-ssm-tools",
    version=ssm_tools.__version__,
    packages=find_packages(),
    scripts=SCRIPTS+[
        'ssm-tunnel-agent'
    ],

    python_requires='>=3.6',

    install_requires=[
        'botocore',
        'boto3',
        'pexpect',
    ],

    package_data={
        '': ['*.txt', '*.md'],
    },

    author="Michael Ludvig",
    author_email="mludvig@logix.net.nz",
    description="Tools for AWS Systems Manager: "+" ".join(SCRIPTS),
    long_description=README,
    long_description_content_type="text/markdown",
    license="Apache License 2.0",
    keywords="aws ssm "+" ".join(SCRIPTS),
    url="https://github.com/mludvig/aws-ssm-tools",
    project_urls={
        "Bug Tracker": "https://github.com/mludvig/aws-ssm-tools/issues",
        "Documentation": "https://github.com/mludvig/aws-ssm-tools/blob/master/README.md",
        "Source Code": "https://github.com/mludvig/aws-ssm-tools",
    }
)
